#ifndef LATTICE_BOLTZMANN_UNITS
#define LATTICE_BOLTZMANN_UNITS

#include <fstream>

class LBMUnitsConverter
{
public:
	LBMUnitsConverter(float u, float Re, int N, int size_x, int size_y, int size_z)
	{
		this->u = u;
		this->Re = Re;
		this->N = N;
		this->size_x = size_x;
		this->size_y = size_y;
		this->size_z = size_z;
	}
	
	// velocity in lattice units
	float getLatticeU() const 
	{
		return u;	
	}
	
	// Reynolds number
	float getRe() const 
	{
		return Re;	
	}

	// resolution
	int getResolution() const 
	{
		return N;	
	}

	// x-length in dimensionless units
	int getSizeX() const 
	{
		return size_x;	
	}

	// y-length in dimensionless units
	int getSizeY() const 
	{
		return size_y;	
	}

	// z-length in dimensionless units
	int getSizeZ() const 
	{
		return size_z;	
	}

	// lattice spacing in dimensionless units
    float getDeltaX() const  
	{ 
		return  (float) 1.0f/ (float) getResolution(); 
	}

    // time step in dimensionless units
    float getDeltaT() const
	{
		return getDeltaX()*getLatticeU(); 
	}

    // conversion from dimensionless to lattice units for space coordinate
    float nCell(int l) const 
	{
		return (int)(l/getDeltaX()+(int)0.5); 
	}

    // conversion from dimensionless to lattice units for time coordinate
    float nStep(int t) const 
	{ 
		return (int)(t/getDeltaT()+(int)0.5);
	}

    // number of lattice cells in x-direction
    int getNx() const    
	{ 
		return nCell(size_x); 
	}

    // number of lattice cells in y-direction
    int getNy() const 
	{
		return nCell(size_y);
	}

    // number of lattice cells in z-direction
    int getNz() const   
	{ 
		return nCell(size_z); 
	} 

    // viscosity in lattice units
    float getLatticeNu() const   
	{ 
		return getLatticeU()*getResolution()/Re;
	}

    // relaxation time
    float getTau() const   
	{
		return 3*getLatticeNu()+0.5f; 
	}

    // relaxation frequency
    float getOmega() const
	{ 
		return 1 / getTau();
	}

	void printValuesConsole()
	{
		printf("Velocity in lattice units: u=%f\n", getLatticeU());
		printf("Reynolds number: Re=%f\n", getRe());
		printf("Lattice resolution: N=%d\n", getResolution());
		printf("Extent of the system: x=%d\n", getSizeX());
		printf("Extent of the system: y=%d\n", getSizeY());
		printf("Extent of the system: z=%d\n", getSizeZ());
		printf("Grid spacing deltaX: dx=%f\n", getDeltaX());
		printf("Time step deltaT: dt=%f\n", getDeltaT());
		printf("Number of lattice cells in: x=%d\n", getNx());
		printf("Number of lattice cells in: y=%d\n", getNy());
		printf("Number of lattice cells in: z=%d\n", getNz());
		printf("Relaxation time: tau=%f\n", getTau());
		printf("Relaxation frequency: omega=%f\n", getOmega());
		printf("Viscosity in lattice units: nu=%f\n", getLatticeNu());	
		printf("\n");
	}

private:
	float u, Re;
	int N, size_x, size_y, size_z;
};

#endif 
