// String objects are a special type of container, specifically designed to operate with sequences of characters
#include <string> 

// Interface to read and write data from files as input/output streams
#include <fstream>

// The OpenGL Extension Wrangler Library 
#include <GL/glew.h>

// The OpenGL Utility Toolkit
#include <GL/glut.h>


//OpenCL
#include "CL/cl.h"

// The lattice Boltzmann method
#include "LBM_3D.h"
//#include "LBMUnitsConverter.h"

#include "Simulation_threaded.h"
#include "Simulation_mpi.h"
#include "Simulation_utils.h"



using namespace std;


extern double MLN;

//Timings:
extern double time_total;
extern double time_loaddata;
extern double time_lbm;
extern double time_setup;
extern double time_wait_to_finish;
extern double time_wait_to_finish_comm;
extern double time_reduce;

extern real_t porosity;

extern long long int voxelCount;
extern long long int ActivevoxelCount;

//only used in the validation
void createHostDataStructureAndCalculateTotalAndActiveVoxels(int *solid_h, uchar *data, int ghostLayersNorth, int ghostLayersSouth);
void Validate(int ant_threads);
void writeForcesToFile(int ant_threads);
extern uchar *data;
extern int iter_count;


int mpi_rank = 0;
int mpi_size = 0;

int main(int argc, char** argv)
{



#if MODE == mode_mpi || MODE == mode_hybrid    //only get rank and size when actually running in MPI mode
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &mpi_size);
#endif

    
  int ant_threads = 1;  
  if (argc > 1){ 
      ant_threads = atoi(argv[1]);
  }


  if (mpi_rank == 0){ //As mpi_rank is set to 0 this will also run when in thread mode
    
    long t1;
    long t2;


    //Start timer for the load data part:
    getTimeInMs(t1);

	cout << "MODE: " << MODE << endl;
    cout << "Loading stone " << DATA_FILE << endl;
    if (DATASET == 4){  //this is the validation set
      createPoiseuilleFlow1D(32,32,32);
    }else if (DATASET >= 9){
      int latticeSize = (DATASET - 8) * 32;
      createPoiseuilleFlow1D(latticeSize,latticeSize,latticeSize);
    }else{
      loadData(EXTRA_LAYERS);
    }

    
    
    cout << "Done loding data" << endl;
    MLN = ActivevoxelCount / 1.0e6;

    //callculate the porosity
    porosity = getPorosity();

    
    
#if PERIODIC_Z
    //If peropdic in z direction we have to add 2 layer to have ghost layer in north for 0 and south for ant - 1
    //This will change the size of the data without changing the getSizeZ    
    extendData(1,1);
#endif

    //Stop timer for the load data
    getTimeInMs(t2);
    time_loaddata = (double (t2 - t1) / toSec);
  }

  
  
  
#if MODE == mode_mpi || MODE == mode_hybrid  
    runSimulationMPI(mpi_rank,mpi_size, N_THREADS_PER_NODE);
#if DATASET == 4 //this is the validation set
    Validate(mpi_size); 
    //writeForcesToFile(mpi_size);
#endif

#elif MODE == mode_thread 
    runSimulationThreaded(ant_threads);
#if DATASET == 4 //this is the validation set
    //writeForcesToFile(mpi_size);
    Validate(ant_threads);   
#endif
#else

#endif

    if(mpi_rank == 0){
      printFinaleState(stdout);
    }
#if DEBUG

    //Print indevidual timings
#if MODE == mode_mpi || MODE == mode_hybrid

  int ant = 2000;
  char * buf = (char *) malloc(ant * sizeof(char));
  char * rbuf = (char *) malloc(ant * sizeof(char));
  MPI_Request request;

  for( int i = 0; i < mpi_size; ++i ) {
	  if (i == 0 && mpi_rank == 0){
		  fprintf(stdout,"\n\nTimings: \n");
	  }

	  if ( i == mpi_rank ) {
		  printIndevidualTimingsMPI(buf,N_THREADS_PER_NODE);
		  MPI_Isend(buf,ant,MPI_CHAR,0,i,MPI_COMM_WORLD,&request);
	  }

	  if (mpi_rank == 0){
		  MPI_Recv(rbuf,ant,MPI_CHAR,i,i,MPI_COMM_WORLD,MPI_STATUSES_IGNORE);
		  printf("%s",rbuf);
	  }
  }

  free(buf);
  free(rbuf);
#elif MODE == mode_thread
    printf("\n\nTimings: \n");
    printIndevidualTimingsThreaded(ant_threads);

#endif 

	if(mpi_rank == 0){

		fprintf(stdout,"\n\nTotale: \n");
		fprintf(stdout,"Setup\t\t\t\t %f\n", time_setup); 
		fprintf(stdout,"Load data\t\t\t %f\n", time_loaddata); 
		fprintf(stdout,"Wait to finish compute\t\t %f\n", time_wait_to_finish); 
		fprintf(stdout,"Wait to finish communicate\t %f\n", time_wait_to_finish_comm); 
		fprintf(stdout,"Reduce\t\t\t\t %f\n", time_reduce); 
		fprintf(stdout,"LBM\t\t\t\t %f\n", time_lbm); 
		fprintf(stdout,"TOTALE TIME:\t\t\t %f\n", time_total ); 
		fprintf(stdout,"\n\n");

	}
#endif

  if(mpi_rank == 0){
  
    std::string file_n;
    file_n += LOGFOLDER;
    
    file_n += "Benchmark";

    stringstream out;
    out << "MODE_" << MODE;
    file_n += "_" + out.str();

    FILE *pFile;
    pFile = fopen(file_n.c_str() ,"a+");


    fprintf(pFile,"%d %e %e\n",getSizeX(),time_lbm, (iter_count * MLN) / time_lbm);
    
    
    fclose(pFile);

  }


#if MODE == mode_mpi || MODE == mode_hybrid 
    MPI_Finalize();
#endif

  return 0;
}





void Validate(int ant_threads)
{
  real_t *u_new;
  int *h_solids;
  if(mpi_rank == 0){ 
    //Validation via PoiseuilleFlow2D
    u_new = new real_t[ActivevoxelCount];
    h_solids = new int[getSizeX()*getSizeY()*getSizeZ()];
    
    createHostDataStructureAndCalculateTotalAndActiveVoxels(h_solids,data+(getSizeX() * getSizeY()),0,0);
  }


    //coppy data back to CPU  
#if MODE == mode_mpi || MODE == mode_hybrid   
  copyAvgVelToCPU_MPI(u_new,mpi_size,N_THREADS_PER_NODE);
#elif MODE == mode_thread
  copyAvgVelToCPU_threaded(u_new,ant_threads);
#endif

    if (mpi_rank == 0){   

      FILE *pFile;
      std::string fName;
      fName = LOGFOLDER; 
      fName += "PoiseuilleFlow2D.txt";
      pFile = fopen(fName.c_str(),"w+");
      writePoiseuilleFlow2DToFile(pFile,h_solids,u_new);

      delete[] u_new;
      delete[] h_solids;
      //fclose(velFile);
      fclose(pFile);
    }

}

void writeForcesToFile(int ant_threads)
{
  real_t *u_new;
  int *h_solids;
  if(mpi_rank == 0){ 
    //Validation via PoiseuilleFlow2D
    u_new = new real_t[ActivevoxelCount];
    h_solids = new int[getSizeX()*getSizeY()*getSizeZ()];
    
    createHostDataStructureAndCalculateTotalAndActiveVoxels(h_solids,data,0,0);
    //createHostDataStructureAndCalculateTotalAndActiveVoxels(h_solids,data+(getSizeX() * getSizeY()),0,0);
  }


    //coppy data back to CPU  
#if MODE == mode_mpi || MODE == mode_hybrid  
  copyAvgVelToCPU_MPI(u_new,mpi_size,N_THREADS_PER_NODE);
#elif MODE == mode_thread 
  copyAvgVelToCPU_threaded(u_new,ant_threads);
#endif

    if (mpi_rank == 0){   
      
      int index = 0;
      FILE *pFile;
      std::string fName;
      fName = LOGFOLDER; 
      fName += "Forces.txt";
      pFile = fopen(fName.c_str(),"w+");
   

    float deltaX = 0.0;

      for (int z = 16;z<17;z++){
        for (int y = 0;y<32;y++){
          for (int x = 0;x<32;x++){
            //fprintf(velFile,"%e ",u_new[z * (32*32) + y * 32 + x]);
            index = h_solids[z * (32*32) + y * 32 + x];
            if( index == -1)
              deltaX = 0.0;
            else
              deltaX = u_new[index] * 1000.0;
      
            fprintf(pFile,"%e %e %e %e\n",(double) x, (double) y ,(double) deltaX ,0.0);
          }
          //fprintf(velFile,"\n");
        }
        //fprintf(velFile,"\n");
      }



      delete[] u_new;
      delete[] h_solids;
      //fclose(velFile);
      fclose(pFile);
    }

}

void createHostDataStructureAndCalculateTotalAndActiveVoxels(int *solid_h, uchar *data, int ghostLayersNorth, int ghostLayersSouth)
{

  //int zEnd =  getSizeZ()+ghostLayersNorth + ghostLayersSouth;
  int voxelIndex = 0;
  int index = 0;
  int obsIndex = 0;
  //total_voxels = 0;
  //voxelsActive = 0;

  for(int z= 0 ; z< getSizeZ(); z++)
  {
    for(int y=0; y<getSizeY(); y++)
    {
      for(int x=0; x<getSizeX(); x++)
      {	
        if(data[index] == 0){
          obsIndex = voxelIndex++;

        }else{
          obsIndex = -1;
        }
        solid_h[index] = obsIndex;
        index++;    
      }
    }
  }
}
